var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { SquareDrawer } from "./ShapeDrawers/SquareDrawer";
import { TextDrawer } from "./ShapeDrawers/TextDrawer";
import { ImageDrawer } from "./ShapeDrawers/ImageDrawer";
import { Plugins } from "./Utils";
import { LineDrawer } from "./ShapeDrawers/LineDrawer";
import { CircleDrawer } from "./ShapeDrawers/CircleDrawer";
import { TriangleDrawer } from "./ShapeDrawers/TriangleDrawer";
import { StarDrawer } from "./ShapeDrawers/StarDrawer";
import { PolygonDrawer } from "./ShapeDrawers/PolygonDrawer";
import { Loader } from "./Core/Loader";
import { Bouncer } from "./Interactions/External/Bouncer";
import { Bubbler } from "./Interactions/External/Bubbler";
import { Connector } from "./Interactions/External/Connector";
import { Grabber } from "./Interactions/External/Grabber";
import { Attractor as MouseAttractor } from "./Interactions/External/Attractor";
import { Repulser as MouseRepulser } from "./Interactions/External/Repulser";
import { Attractor as ParticlesAttractor } from "./Interactions/Particles/Attractor";
import { Collider } from "./Interactions/Particles/Collider";
import { Infecter } from "./Interactions/Particles/Infecter";
import { Repulser } from "./Interactions/Particles/Repulser";
import { Linker } from "./Interactions/Particles/Linker";
import { LifeUpdater } from "./Updaters/LifeUpdater";
import { OpacityUpdater } from "./Updaters/OpacityUpdater";
import { SizeUpdater } from "./Updaters/SizeUpdater";
import { AngleUpdater } from "./Updaters/AngleUpdater";
import { ColorUpdater } from "./Updaters/ColorUpdater";
import { StrokeColorUpdater } from "./Updaters/StrokeColorUpdater";
import { OutOfCanvasUpdater } from "./Updaters/OutOfCanvasUpdater";
export class MainSlim {
    constructor() {
        this.initialized = false;
        const squareDrawer = new SquareDrawer();
        const textDrawer = new TextDrawer();
        const imageDrawer = new ImageDrawer();
        Plugins.addParticleUpdater((container) => (particle) => new LifeUpdater(container, particle));
        Plugins.addParticleUpdater((container) => (particle) => new OpacityUpdater(container, particle));
        Plugins.addParticleUpdater((container) => (particle) => new SizeUpdater(container, particle));
        Plugins.addParticleUpdater((container) => (particle) => new AngleUpdater(container, particle));
        Plugins.addParticleUpdater((container) => (particle) => new ColorUpdater(container, particle));
        Plugins.addParticleUpdater((container) => (particle) => new StrokeColorUpdater(container, particle));
        Plugins.addParticleUpdater((container) => (particle) => new OutOfCanvasUpdater(container, particle));
        Plugins.addInteractor((container) => new Bouncer(container));
        Plugins.addInteractor((container) => new Bubbler(container));
        Plugins.addInteractor((container) => new Connector(container));
        Plugins.addInteractor((container) => new Grabber(container));
        Plugins.addInteractor((container) => new MouseAttractor(container));
        Plugins.addInteractor((container) => new MouseRepulser(container));
        Plugins.addInteractor((container) => new ParticlesAttractor(container));
        Plugins.addInteractor((container) => new Collider(container));
        Plugins.addInteractor((container) => new Infecter(container));
        Plugins.addInteractor((container) => new Repulser(container));
        Plugins.addInteractor((container) => new Linker(container));
        Plugins.addShapeDrawer("line", new LineDrawer());
        Plugins.addShapeDrawer("circle", new CircleDrawer());
        Plugins.addShapeDrawer("edge", squareDrawer);
        Plugins.addShapeDrawer("square", squareDrawer);
        Plugins.addShapeDrawer("triangle", new TriangleDrawer());
        Plugins.addShapeDrawer("star", new StarDrawer());
        Plugins.addShapeDrawer("polygon", new PolygonDrawer());
        Plugins.addShapeDrawer("char", textDrawer);
        Plugins.addShapeDrawer("character", textDrawer);
        Plugins.addShapeDrawer("image", imageDrawer);
        Plugins.addShapeDrawer("images", imageDrawer);
    }
    init() {
        if (!this.initialized) {
            this.initialized = true;
        }
    }
    loadFromArray(tagId, options, index) {
        return __awaiter(this, void 0, void 0, function* () {
            return Loader.load(tagId, options, index);
        });
    }
    load(tagId, options) {
        return __awaiter(this, void 0, void 0, function* () {
            return Loader.load(tagId, options);
        });
    }
    set(id, element, options) {
        return __awaiter(this, void 0, void 0, function* () {
            return Loader.set(id, element, options);
        });
    }
    loadJSON(tagId, pathConfigJson, index) {
        return Loader.loadJSON(tagId, pathConfigJson, index);
    }
    setOnClickHandler(callback) {
        Loader.setOnClickHandler(callback);
    }
    dom() {
        return Loader.dom();
    }
    domItem(index) {
        return Loader.domItem(index);
    }
    addShape(shape, drawer, init, afterEffect, destroy) {
        let customDrawer;
        if (typeof drawer === "function") {
            customDrawer = {
                afterEffect: afterEffect,
                destroy: destroy,
                draw: drawer,
                init: init,
            };
        }
        else {
            customDrawer = drawer;
        }
        Plugins.addShapeDrawer(shape, customDrawer);
    }
    addPreset(preset, options) {
        Plugins.addPreset(preset, options);
    }
    addPlugin(plugin) {
        Plugins.addPlugin(plugin);
    }
    addNoiseGenerator(name, generator) {
        Plugins.addNoiseGenerator(name, generator);
    }
    addInteractor(interactorInitializer) {
        Plugins.addInteractor(interactorInitializer);
    }
    addParticleUpdater(updaterInitializer) {
        Plugins.addParticleUpdater(updaterInitializer);
    }
}
